# Copyright 2021-2024 KU Leuven.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Rinaldo Wander Montalvão, PhD
#
import math
import warnings

import numpy as np

from sty import fg

from numpy.random import rand, randint

from Bio.PDB import PDBParser, PDBIO
from Bio.SeqUtils import seq1
from Bio.SVDSuperimposer import SVDSuperimposer
from Bio.PDB.PDBExceptions import PDBConstructionWarning

from sklearn.preprocessing import StandardScaler
from sklearn.cluster import AgglomerativeClustering

warnings.filterwarnings('ignore', category=PDBConstructionWarning)


def rmsd(x: np.ndarray, y: np.ndarray) -> float:
    """
         Compute the RMSD between the two numpy arrays.
         :param x: The first numpy array
         :type x: np.ndarray
         :param y: The second numpy array
         :type y: np.ndarray
         :return: The RMSD between the two numpy arrays
         :rtype: float
    """

    return np.sqrt((((x - y) ** 2) * 3).mean())


def select(data: np.ndarray, seg: list, msk: list) -> np.ndarray:
    """
    Select the data in the given segment from the mask.
    :param data: The data to select
    :type data: np.ndarray
    :param seg: The segment to select
    :type seg: list
    :param msk: The mask for aligned segment
    :type msk: list
    :return: The selected data
    :rtype: np.ndarray
    """
    tmp = []
    for i, j in enumerate(seg):
        if msk[i]:
            tmp.append(data[j])

    return np.array(tmp)


def superposition(xo: np.ndarray, yo: np.ndarray, seg: list, msk: list) -> tuple[np.ndarray, np.ndarray]:
    """
    Compute the superposition of the two numpy arrays.
    :param xo: The first numpy array
    :type xo: np.ndarray
    :param yo: The second numpy array
    :type yo: np.ndarray
    :param seg: The segment to superimpose
    :type seg: list
    :param msk: The mask for the aligned segment
    :type msk: list
    :return: The superposition of the two numpy arrays
    :rtype: tuple(np.ndarray, np.ndarray)
    """
    x = select(xo, seg, msk)
    y = select(yo, seg, msk)

    sup = SVDSuperimposer()

    sup.set(x, y)
    sup.run()

    return sup.get_rotran()


def energy(xo: np.ndarray, yo: np.ndarray, seg: list, msk: list) -> float:
    """
    Compute the energy between two numpy arrays.
    :param xo: The first numpy array
    :type xo: np.ndarray
    :param yo: The second numpy array
    :type yo: np.ndarray
    :param seg: The segment to superimpose
    :type seg: list
    :param msk: The mask for the aligned segment
    :type msk: list
    :return: The energy between two numpy arrays
    :rtype: float
    """
    rot, tran = superposition(xo, yo, seg, msk)

    yt = np.dot(yo, rot) + tran

    return rmsd(xo, yt)


def segments(anchors: list, members: list) -> list:
    """
    Return a list of segments for each member of the protein alignment of an anchor.
    :param anchors: The list of anchors
    :type anchors: list
    :param members: The list of members
    :type members: list
    :return: The of segments
    :rtype: list
    """
    seg = []
    for member in members:
        ini, end = anchors[member]
        seg.extend([x for x in range(ini, end)])

    return seg


def simulated_annealing(xo: np.ndarray,
                        yo: np.ndarray,
                        anchors: list,
                        members: list) -> tuple[np.ndarray, np.ndarray, float]:
    """
    Compute the rotation and translation matrices that reduces the RMSD
    between two protein segments.
    :param xo: The first numpy array
    :type xo: np.ndarray
    :param yo: The second numpy array
    :type yo: np.ndarray
    :param anchors: The list of anchors
    :type anchors: list
    :param members: The list of members
    :type members: list
    :return: The rotation and translation matrices and the final RMSD
    :rtype: tuple(np.ndarray, np.ndarray, float)
    """
    seg = segments(anchors, members)

    msk0 = [True for _ in range(len(seg))]

    temperature = 300
    energy0 = energy(xo, yo, seg, msk0)
    while temperature > 0.00001:
        for i in range(10000):
            msk1 = msk0.copy()
            j = randint(len(msk1))
            if msk1[j]:
                msk1[j] = False
            else:
                msk1[j] = True

            energy1 = energy(xo, yo, seg, msk1)

            energy_delta = energy1 - energy0

            if energy_delta < 0.0 or rand() < math.exp(-energy_delta / temperature):
                msk0 = msk1.copy()
                energy0 = energy1

        # TODO: test other schedules
        temperature *= 0.1

    rot, tran = superposition(xo, yo, seg, msk0)
    return rot, tran, energy0


def superimposer(align, threshold=0.8, csv=False, pdb=True):
    """
    Given an alignment, superimpose the PDB structure using Differential Geometry
    as dissimilarity measure. It can output CSV files with the alpha carbon coordinates
    and cluster groups, and also save the superimposed PDB structures.
    :param align: Alignment to superimpose
    :type align: Bio.Align.MultipleSeqAlignment
    :param threshold: Threshold for cluster selection
    :type threshold: float
    :param csv: If true, output the CSV files
    :type csv: bool
    :param pdb: If true, output the superimposed PDB structures
    :type pdb: bool
    """
    # Collect Differential Geometry data for initial clustering
    data = []
    id2pos = {}

    for position, record in enumerate(align):
        if 'structure' in record.description:
            id2pos[record.id] = position
            pairs = zip(
                record.letter_annotations['curvature'],
                record.letter_annotations['torsion'],
            )
            data = [list(pair) for pair in pairs]
            record.letter_annotations['cluster'] = [0 for _ in record.seq]

    scaler = StandardScaler()
    scaler.fit(data)

    # Clustering to find maximal conserved regions
    clustering = AgglomerativeClustering(distance_threshold=threshold, n_clusters=None)
    for i in range(align.get_alignment_length()):
        xy = []
        tags = []
        for identity, position in id2pos.items():
            record = align[position]
            if record.seq[i] != '-':
                xy.append(
                    [
                        record.letter_annotations['curvature'][i],
                        record.letter_annotations['torsion'][i],
                    ]
                )
                tags.append(identity)

        if len(xy) > 1:
            xy = scaler.transform(xy)

            clusters = clustering.fit_predict(xy)

            map_of_clusters = {pair[0]: pair[1] for pair in zip(tags, clusters)}

            for identity, position in id2pos.items():
                record = align[position]
                if identity in map_of_clusters:
                    record.letter_annotations['cluster'][i] = (
                            map_of_clusters[identity] + 1
                    )
                    # print(id, record.letter_annotations['cluster'][i])
        else:
            for identity, position in id2pos.items():
                record = align[position]
                record.letter_annotations['cluster'][i] = 0

    # Find the highly conserved areas
    ini = 0
    end = align.get_alignment_length()

    area = []
    for i in range(ini, end):
        s = set()
        for identity, position in id2pos.items():
            record = align[position]
            j = record.letter_annotations['cluster'][i]
            s = s.union({j})
        # Only regions conserved across all proteins
        if len(s) == 1:
            area.append(1)
        else:
            area.append(0)

    # Find all anchor conserved regions
    anchors = []
    anchor_lengths = []
    new_area = [0 for _ in area]

    ini, end = 0, 0
    while ini < len(area):
        while ini < len(area):
            if area[ini] == 1:
                break
            else:
                ini += 1

        end = ini
        while end < len(area):
            if area[end] != 1:
                break
            else:
                end += 1

        # Minimal length for an anchor regions is 5 residues long
        if end - ini >= 5:
            anchor_lengths.append(end - ini)
            anchors.append((ini, end))
            for i in range(ini, end):
                new_area[i] = 1
        ini = end

    # Group is the annotation for the new clusters
    for identity, position in id2pos.items():
        record = align[position]
        record.letter_annotations['group'] = [0 for _ in record.seq]
        record.letter_annotations['ca_coords'] = [[0.0, 0.0, 0.0] for _ in record.seq]

    for anchor in anchors:
        for identity, position in id2pos.items():
            record = align[position]
            ini, end = anchor
            for i in range(ini, end):
                record.letter_annotations['group'][i] = 1

    # Select the top 5 anchor regions
    top_anchors = np.argsort(anchor_lengths)[::-1]
    if len(top_anchors) > 5:
        top_anchors = top_anchors[:5]

    # Read the CA coordinates for all the files
    parser = PDBParser()

    ca_coords = {}
    ca_masked = {}
    structures = {}
    for record in align:
        if 'structure' not in record.description:
            continue
        id = record.id
        structures[id] = parser.get_structure(id, f'{id}.pdb')

        model = structures[id][0]
        xyz = []
        seq = []
        for chain in model:
            for residue in chain:
                xyz.append(residue['CA'].get_coord())
                seq.append(seq1(residue.get_resname()))
        j = 0
        cds = []
        msk = []
        for i in range(align.get_alignment_length()):
            # Insure coordinate alignment
            if record.seq[i] == '-':
                cds.append(np.array([0.0, 0.0, 0.0]))
                msk.append(False)
            elif record.seq[i] == seq[j]:
                cds.append(xyz[j])
                msk.append(True)
                j += 1
            else:
                print(f'Error: {i} {record.seq[i]} - {j} {seq[j]}')
        ca_coords[id] = np.array(cds)
        ca_masked[id] = np.array(msk)

    # Optimise the anchor positions for minimal RMSD
    # TODO: Use a dendrogram for the order?
    ids = list(ca_coords.keys())
    ref = ids.pop(0)
    print(f'{ref}')

    xo = ca_coords[ref]
    for idx in ids:
        yo = ca_coords[idx]
        rot, tran, E0 = simulated_annealing(xo, yo, anchors, top_anchors)
        ca_coords[idx] = np.dot(yo, rot) + tran
        print(f'{idx}: {E0:.2f} Å')

        for model in structures[idx]:
            for chain in model:
                for residue in chain:
                    for atom in residue:
                        ya = atom.get_coord()
                        atom.set_coord(np.dot(ya, rot) + tran)

    # Store the CA coordinates in the record
    for idx in ca_coords.keys():
        for i, m in enumerate(ca_masked[idx]):
            if m:
                record = align[id2pos[idx]]
                x = ca_coords[idx][i][0]
                y = ca_coords[idx][i][1]
                z = ca_coords[idx][i][2]
                record.letter_annotations['ca_coords'][i] = [x, y, z]

    # Remove group outliers
    for anchor in anchors:
        ini, end = anchor
        for i in range(ini, end):
            max_dist = 0.0
            for j in id2pos.keys():
                cj = ca_coords[j][i]
                for k in id2pos.keys():
                    if k != j:
                        ck = ca_coords[k][i]
                        dist = np.linalg.norm(ck - cj)
                        if dist > max_dist:
                            max_dist = dist

            # Any region with at least one distance
            # over the threshold is removed
            if max_dist > 2.0:
                for j in id2pos.values():
                    record = align[j]
                    record.letter_annotations['group'][i] = 0

    # Remove conserved regions with less than 3 residues
    ini = 0
    end = align.get_alignment_length()

    area = []
    for i in range(ini, end):
        s = set()
        for position in id2pos.values():
            record = align[position]
            j = record.letter_annotations['group'][i]
            if j == 1:
                s = s.union({j})
        if len(s) == 1:
            area.append(1)
        else:
            area.append(0)

    regions = []
    regions_lengths = []
    new_area = [0 for _ in area]

    ini, end = 0, 0
    while ini < len(area):
        while ini < len(area):
            if area[ini] == 1:
                break
            else:
                ini += 1

        end = ini
        while end < len(area):
            if area[end] != 1:
                break
            else:
                end += 1
        delta = end - ini
        if 0 < delta < 3:
            regions_lengths.append(delta)
            regions.append((ini, end))
            for i in range(ini, end):
                new_area[i] = 1
        ini = end

    for region in regions:
        ini, end = region
        for i in range(ini, end):
            for position in id2pos.values():
                record = align[position]
                record.letter_annotations['group'][i] = 0

    # Remove non-conserved regions with less than 3 residues
    ini = 0
    end = align.get_alignment_length()

    area = []
    for i in range(ini, end):
        s = set()
        for position in id2pos.values():
            record = align[position]
            j = record.letter_annotations['group'][i]
            if j == 0:
                s = s.union({j})
        if len(s) == 1:
            area.append(1)
        else:
            area.append(0)

    regions = []
    regions_lengths = []
    new_area = [0 for _ in area]

    ini, end = 0, 0
    while ini < len(area):
        while ini < len(area):
            if area[ini] == 1:
                break
            else:
                ini += 1

        end = ini
        while end < len(area):
            if area[end] != 1:
                break
            else:
                end += 1
        delta = end - ini
        if 0 < delta < 3:
            regions_lengths.append(delta)
            regions.append((ini, end))
            for i in range(ini, end):
                new_area[i] = 1
        ini = end

    for region in regions:
        ini, end = region
        for i in range(ini, end):
            for position in id2pos.values():
                record = align[position]
                record.letter_annotations['group'][i] = 1

    # Output CSV coordinate files
    if csv:
        for idx in ca_coords.keys():
            record = align[id2pos[idx]]
            with open(f'{idx}.csv', 'w') as f:
                for i, m in enumerate(ca_masked[idx]):
                    if m:
                        x = ca_coords[idx][i][0]
                        y = ca_coords[idx][i][1]
                        z = ca_coords[idx][i][2]
                        g = record.letter_annotations['group'][i]
                        f.write(f'{i},{x:.4f},{y:.4f},{z:.4f},{g}\n')

    # Output superposed PDB files
    if pdb:
        io = PDBIO()
        for idx in ca_coords.keys():
            io.set_structure(structures[idx])
            io.save(f'{idx}_sup.pdb')

    return


def show_align(align, pal):
    """
    Show the alignment colored by cluster using a Seaborn pallet
    :param align: Alignment to show
    :type align: Bio.Align.MultipleSeqAlignment
    :param pal: Palette
    :type pal: seaborn.palettes._ColorPalette
    """
    length = align.get_alignment_length()
    total = int(length / 50.0)
    for j in range(total + 1):
        ini = 0 + 50 * j
        end = 49 + 50 * j + 1

        if end >= length:
            end = length - 1

        if end > 100:
            print('      ', end='')
            for i in range(ini, end):
                s = f'{i:03n}'
                if i % 10 == 0:
                    print(s[0], end='')
                else:
                    print(' ', end='')
            print()

        print('      ', end='')
        for i in range(ini, end):
            s = f'{i:03n}'
            if i % 10 == 0:
                print(s[1], end='')
            else:
                print(' ', end='')
        print()

        print('      ', end='')
        for i in range(ini, end):
            s = f'{i:03n}'
            if i % 10 == 0:
                print(s[2], end='')
            else:
                print(' ', end='')
        print()

        for record in align:
            if 'structure' not in record.description:
                continue

            print(record.id, end=':')
            for i in range(ini, end):
                j = record.letter_annotations['group'][i] - 1
                if j >= 0:
                    R = int(pal[j][0] * 255)
                    G = int(pal[j][1] * 255)
                    B = int(pal[j][2] * 255)
                    print(fg(R, G, B) + record.seq[i], end='')
                else:
                    print(fg(125, 125, 125) + record.seq[i].lower(), end='')
            print(fg.rs)
    return



# Copyright 2021-2024 KU Leuven.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Rinaldo Wander Montalvão, PhD
#
import os
from typing_extensions import SupportsIndex

import Bio.Align

import pandas as pd
import nglview as nv
import seaborn as sns

from typing import Dict, Tuple, List, Any
from ipywidgets import Box

from Bio import AlignIO
from Bio.PDB import PDBParser
from Bio.PDB.Structure import Structure

from melodia_py.geometryparser import GeometryParser

from sklearn.preprocessing import StandardScaler
from sklearn.cluster import AgglomerativeClustering

from importlib import resources as importlib_resources

def geometry_from_structure_file(file_name: str) -> pd.DataFrame:
    """
         Function used to compute the geometric properties around residues
         for (BioPython compatible) structures.
         It computes curvature, torsion, arc-length and writhing number
         :param file_name: Protein file name
         :type file_name: String
         :return: Pandas dataframe
         :rtype: pd.DataFrame
    """
    parser = PDBParser()
    name, ext = os.path.splitext(file_name)
    structure = parser.get_structure(name, file_name)

    base = proc_chains(structure)

    return base


def geometry_from_structure(structure: Structure) -> pd.DataFrame:
    """
         Function used to compute the geometric properties around residues
         for (BioPython compatible) structures.
         It computes curvature, torsion, arc-length and writhing number
         :param structure: BioPython PDB Structure class
         :type structure: Structure
         :return: Panda dataframe
         :rtype: pd.DataFrame
    """
    base = proc_chains(structure)

    return base


def proc_chains(structure: Structure) -> pd.DataFrame:
    """
         Function used to compute the geometric properties around residues
         for (BioPython compatible) structures.
         It computes curvature, torsion, arc-length and writhing number
         :param structure: BioPython PDB Structure class
         :type structure: Structure
         :return: Panda dataframe
         :rtype: pd.DataFrame
    """
    pdb_code = structure.id.upper()
    chains = {}
    for model in structure:
        for chain in model:
            chain_id = chain.id
            key = f'{pdb_code}:{model.id}:{chain_id}'
            chains[key] = GeometryParser(chain)

    # Generate a Pandas Dataframe
    identity = []
    model = []
    chain = []
    order = []
    name = []
    curvature = []
    torsion = []
    arc_length = []
    writhing = []
    angle_phi = []
    angle_psi = []
    for key, value in chains.items():
        # print(key)
        tokens = key.split(':')

        for res_id in value.residues:
            model.append(int(tokens[1]))
            chain.append(tokens[2])

            identity.append(value.residues[res_id].res_num)
            order.append(value.residues[res_id].res_order)
            name.append(value.residues[res_id].name)
            curvature.append(value.residues[res_id].curvature)
            torsion.append(value.residues[res_id].torsion)
            arc_length.append(value.residues[res_id].arc_len)
            writhing.append(value.residues[res_id].writhing)
            angle_phi.append(value.residues[res_id].phi)
            angle_psi.append(value.residues[res_id].psi)

    base = {'id': identity,
            'model': model,
            'code': pdb_code,
            'chain': chain,
            'order': order,
            'name': name,
            'curvature': curvature,
            'torsion': torsion,
            'arc_length': arc_length,
            'writhing': writhing,
            'phi': angle_phi,
            'psi': angle_psi}

    return pd.DataFrame(base)


def geometry_dict_from_structure(structure: Structure) -> Dict[str, GeometryParser]:
    """
         Function used to compute the geometric properties around residues
         for (BioPython compatible) structures.
         It computes curvature, torsion, arc-length and writhing number
         :param structure: BioPython PDB Structure class
         :type structure: Structure
         :return: Geometry Dictionary
         :rtype: Dict[str, GeometryParser]
    """
    chains = {}
    for model in structure:
        for chain in model:
            chain_id = chain.id
            key = f'{model.id}:{chain_id}'
            residues = [res.id[1] for res in list(chain.get_residues()) if res.id[0] == ' ']
            if residues:
                chains[key] = GeometryParser(chain)

    return chains


def bfactor_from_geo(structure: Structure, attribute: str, geo: Dict[str, GeometryParser] = None) -> None:
    """
        Set the PDB bfactor to geometric values
         :param structure: BioPython PDB Structure class
         :type structure: Structure
         :param attribute: geometric attribute
         :type attribute: str
        :param geo: Geometry Dictionary
        :type: Dict[str, GeometryParser]

    """
    if geo is None:
        geo = geometry_dict_from_structure(structure)

    # Set all atoms' bfactors to a minimal values
    min_values = []
    for key in geo:
        if attribute == 'curvature':
            min_values.append(min([res.curvature for res in geo[key].residues.values()]))
        elif attribute == 'torsion':
            min_values.append(min([res.torsion for res in geo[key].residues.values()]))
        elif attribute == 'custom':
            min_values.append(min([res.custom for res in geo[key].residues.values()]))
        else:
            min_values.append(0.0)

    min_value = min(min_values)
    for atom in structure.get_atoms():
        if atom.is_disordered():
            for disordered_atom in atom.disordered_get_list():
                disordered_atom.set_bfactor(min_value)
        else:
            atom.set_bfactor(min_value)

    # Change bfactors to a differential geometry value
    for model in structure:
        for chain in model:
            for atom in chain.get_atoms():
                het_flag, sequence_id, insertion_code = atom.get_parent().id
                if het_flag[0] == " ":
                    key = f'{model.id}:{chain.id}'
                    res = geo[key].residues_map[sequence_id]

                    if attribute == 'curvature':
                        value = geo[key].residues[res].curvature
                    elif attribute == 'torsion':
                        value = geo[key].residues[res].torsion
                    elif attribute == 'custom':
                        value = geo[key].residues[res].custom
                    else:
                        value = 0.0

                    atom.set_bfactor(value)


def view_putty(structure: Structure, radius_scale=1.0, width=1200, height=600) -> Box:
    """
        Display PDB structure as a putty model
        :param structure: BioPython PDB Structure class
        :type structure: Structure
        :param radius_scale: Radius Scale
        :type width: float
        :param width: Widget width
        :type width: int
        :param height: Widget height
        :type height: int
    """
    view = nv.show_biopython(structure)
    view.representations = [
        {"type": "tube",
         "params": {
             "sele": "protein",
             "radius": "bfactor",
             "radiusScale": radius_scale,
             "color": "bfactor",
             "colorScale": "RdYlBu"
         }}
    ]
    view.layout.width = '100%'
    view.layout.height = '100%'

    box = Box([view])
    box.layout.width = f'{width}px'
    box.layout.height = f'{height}px'

    return box


def view_cartoon(structure: Structure, width=1200, height=600) -> Box:
    """
        Display PDB structure as a cartoon model
        :param structure: BioPython PDB Structure class
        :type structure: Structure
        :param width: Widget width
        :type width: int
        :param height: Widget height
        :type height: int
    """
    view = nv.show_biopython(structure)
    view.representations = [
        {"type": "cartoon",
         "params": {
             "sele": "protein",
             "color": "bfactor",
             "colorScale": "RdYlBu"
         }}
    ]
    view.layout.width = '100%'
    view.layout.height = '100%'

    box = Box([view])
    box.layout.width = f'{width}px'
    box.layout.height = f'{height}px'

    return box


def view_tube(structure: Structure, width=1200, height=600) -> Box:
    """
        Display PDB structure as a tube model
        :param structure: BioPython PDB Structure class
        :type structure: Structure
        :param width: Widget width
        :type width: int
        :param height: Widget height
        :type height: int
    """
    view = nv.show_biopython(structure)
    view.representations = [
        {"type": "tube",
         "params": {
             "sele": "protein",
             "color": "bfactor",
             "colorScale": "RdYlBu"
         }}
    ]
    view.layout.width = '100%'
    view.layout.height = '100%'

    box = Box([view])
    box.layout.width = f'{width}px'
    box.layout.height = f'{height}px'

    return box


def parser_pir_file(pir_file: str) -> Bio.Align.MultipleSeqAlignment:
    """
        Parser for PIR alignment file and associated PDB structures
        :param pir_file: BioPython PDB Structure class
        :type pir_file: str
        :return: Protein sequence alignment
        :rtype: Bio.Align.MultipleSeqAlignment
    """
    c321 = {'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K',
            'ILE': 'I', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ASN': 'N',
            'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W',
            'ALA': 'A', 'VAL': 'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M'}

    align = AlignIO.read(pir_file, 'pir')

    parser = PDBParser()
    for record in align:
        # Only select structures
        if record.description[:9] != 'structure':
            continue

        file_name = f'{record.id}.pdb'

        structure = parser.get_structure(record.id, file_name)

        # Renumber the residues to start with 1
        residue_number = 1
        for model in structure:
            for chain in model:
                for residue in chain:
                    # print(residue.id)
                    residue.id = (residue.id[0], residue_number, 'Z')
                    residue_number += 1
                for residue in chain:
                    residue.id = (residue.id[0], residue.id[1], ' ')
                    # print('----', residue.id)

        geo = geometry_dict_from_structure(structure)

        curvature = []
        torsion = []
        arc_length = []
        writhing = []
        phi = []
        psi = []

        idx = []
        for key in geo:
            for res in geo[key].residues:
                idx.append((key, res))
        j = 0
        for i, letter in enumerate(record.seq):
            if letter != '-' and letter != '/':
                try:
                    curr_chain = idx[j][0]
                    curr_residue = idx[j][1]
                    res_code = c321[geo[curr_chain].residues[curr_residue].name]
                except KeyError:
                    raise NameError(f'Alignment error: {record.id} {curr_chain} residue {curr_residue + 1} {letter}')
                if letter != res_code:
                    raise NameError(f'Alignment error: {record.id} res={i + 1} seq={letter} pdb={res_code}')
                curvature.append(geo[curr_chain].residues[curr_residue].curvature)
                torsion.append(geo[curr_chain].residues[curr_residue].torsion)
                arc_length.append(geo[curr_chain].residues[curr_residue].arc_len)
                writhing.append(geo[curr_chain].residues[curr_residue].writhing)
                phi.append(geo[curr_chain].residues[curr_residue].phi)
                psi.append(geo[curr_chain].residues[curr_residue].psi)
                j += 1
            else:
                curvature.append(0.0)
                torsion.append(0.0)
                arc_length.append(0.0)
                writhing.append(0.0)
                phi.append(0.0)
                psi.append(0.0)

        record.letter_annotations['curvature'] = curvature
        record.letter_annotations['torsion'] = torsion
        record.letter_annotations['arc_length'] = arc_length
        record.letter_annotations['writhing'] = writhing
        record.letter_annotations['phi'] = phi
        record.letter_annotations['psi'] = psi

    return align


def dataframe_from_alignment(align: Bio.Align.MultipleSeqAlignment, keys=None) -> pd.DataFrame:
    """
         Create a Pandas Dataframe from an annotated alignment
         :param align: Protein alignment
         :type align: Bio.Align.MultipleSeqAlignment
         :param keys: keys for the annotations
         :type keys: list
         :return: Pandas dataframe
         :rtype: pd.DataFrame
    """
    data = {}

    for record in align:
        if record.description[:9] == 'structure':
            data[f'seq_{record.id}'] = list(record.seq)
            if keys is None:
                items = record.letter_annotations.keys()
            else:
                items = keys
            for key in items:
                data[f'{key}_{record.id}'] = record.letter_annotations[key]

    return pd.DataFrame.from_dict(data)


class PropensityTable:
    """
    Class for the Protein Propensity Table data
    """
    __slots__ = ('__ini', '__end', '__lines', '__data')

    def __init__(self):
        """
        Load the Propensity Table data
        """

        def find_block():
            found = False
            while self.__ini < self.__end:
                if self.__lines[self.__ini][0] == '>':
                    found = True
                    break
                self.__ini += 1
            return found

        def load_block():
            tag = self.__lines[self.__ini].split()
            self.__ini += 1
            rows = []
            while self.__lines[self.__ini][0] != 'U':
                rows.append(self.__lines[self.__ini])
                self.__ini += 1
            rows.append(self.__lines[self.__ini])
            return int(tag[1]), rows

        ref = importlib_resources.files('melodia_py').joinpath('data/luthier.dat')
        with ref.open('rb') as fp:
            lines = fp.readlines()

        # Remove newlines
        for i, line in enumerate(lines):
            lines[i] = lines[i].decode('utf-8')
            if lines[i][-1] == '\n':
                lines[i] = lines[i][:-1]

        # TODO: trim data after use
        self.__ini = 0
        self.__end = len(lines)
        self.__lines = lines
        self.__data = {}

        while find_block():
            key, table = load_block()

            dct = {}
            head = table[0].split()
            for i in range(1, len(table)):
                row = table[i].split()
                key1 = row[0]
                for j in range(1, len(row)):
                    key2 = head[j]
                    key12 = f'{key1},{key2}'
                    value = int(row[j])
                    dct[key12] = value

            self.__data[key] = dct

    def get_score(self, target: str, residue: str, phi: float, psi: float) -> int:
        """
        Find anomalies in the protein's chain
        :param target: Target residue 1 letter code
        :type target: str
        :param residue: Residue 1 letter code
        :type residue: str
        :param phi: phi angle (degrees)
        :type phi: float
        :param psi: psi angle (degrees)
        :type psi: float
        :return: Score
        :rtype: int
        """
        max_phi = [0.0, 0.0, 0.0, -110.0, -110.0, 0.0, 140.0, 180.0, 180.0]
        min_phi = [-180.0, -110.0, -110.0, -180.0, -180.0, -180.0, 20.0, 0.0, 0.0]
        max_psi = [45.0, 180.0, -90.0, 180.0, -90.0, 100.0, 80.0, -40.0, 180.0]
        min_psi = [-90.0, 100.0, -180.0, 100.0, -180.0, 45.0, -40.0, -180.0, 80.0]

        tab_map = [0, 1, 1, 2, 2, 3, 4, 5, 5]

        score = 0
        for i in range(len(tab_map)):
            if (min_phi[i] <= phi < max_phi[i]) and (min_psi[i] <= psi < max_psi[i]):
                key = f'{target.upper()},{residue.upper()}'
                score = self.__data[3][key]
                break

        return score


def cluster_alignment(align: Bio.Align.MultipleSeqAlignment, threshold=0.7, long=False) -> None:
    """
         Cluster the alignments by structural similarity
         :param align: Protein alignment
         :type align: Bio.Align.MultipleSeqAlignment
         :param threshold: similarity threshold
         :type threshold: float
         :param long: output only long cluster (length > 3)
         :type long: bool
    """
    data = []
    id2pos = {}
    for position, record in enumerate(align):
        if 'structure' in record.description:
            id2pos[record.id] = position
            for pair in zip(record.letter_annotations['curvature'], record.letter_annotations['torsion']):
                data.append(list(pair))
            record.letter_annotations['cluster'] = [0 for _ in record.seq]

    scaler = StandardScaler()
    scaler.fit(data)

    clustering = AgglomerativeClustering(distance_threshold=threshold, n_clusters=None)
    for i in range(align.get_alignment_length()):
        xy = []
        tags = []
        for identity, position in id2pos.items():
            record = align[position]
            if record.seq[i] != '-':
                xy.append([record.letter_annotations['curvature'][i], record.letter_annotations['torsion'][i]])
                tags.append(identity)

        if len(xy) > 1:
            xy = scaler.transform(xy)

            clusters = clustering.fit_predict(xy)

            map_of_clusters = {pair[0]: pair[1] for pair in zip(tags, clusters)}

            for identity, position in id2pos.items():
                record = align[position]
                if identity in map_of_clusters:
                    record.letter_annotations['cluster'][i] = map_of_clusters[identity]
                    # print(id, record.letter_annotations['cluster'][i])
        else:
            for identity, position in id2pos.items():
                record = align[position]
                record.letter_annotations['cluster'][i] = 0

    last_cluster = 0
    for record in align:
        if 'structure' not in record.description:
            continue
        max_cluster = max(record.letter_annotations['cluster'])
        if max_cluster > last_cluster:
            last_cluster = max_cluster

    for i in range(align.get_alignment_length() - 1):
        j = i + 1
        left = {}
        right = {}
        for k, record in enumerate(align):
            if 'structure' not in record.description:
                continue

            ca = record.letter_annotations['cluster'][i]
            cb = record.letter_annotations['cluster'][j]

            if ca not in left:
                left[ca] = [k]
            else:
                left[ca].append(k)

            if cb not in right:
                right[cb] = [k]
            else:
                right[cb].append(k)

        for key, value in left.items():
            left[key] = set(value)

        for key, value in right.items():
            right[key] = set(value)

        for right_key in right:
            found_key = None
            for left_key in left:
                if right[right_key].symmetric_difference(left[left_key]) == set():
                    found_key = left_key

            if found_key is None:
                last_cluster += 1
                for k in right[right_key]:
                    record = align[k]
                    record.letter_annotations['cluster'][j] = last_cluster
            else:
                for k in right[right_key]:
                    record = align[k]
                    record.letter_annotations['cluster'][j] = found_key

    data, idx = get_idx(align)

    if long:
        last_cluster = 0
        for j in idx:
            cluster, ini, end, size = data[j]
            if size < 3:
                for i in range(ini, end + 1):
                    for record in align:
                        if 'structure' not in record.description:
                            continue
                        if record.letter_annotations['cluster'][i] == cluster:
                            record.letter_annotations['cluster'][i] = -1
                data[j] = (-1, ini, end, size)
            else:
                # print(f'{ini}-{end} {cluster}->{p}')
                for i in range(ini, end + 1):
                    for record in align:
                        if 'structure' not in record.description:
                            continue
                        if record.letter_annotations['cluster'][i] == cluster:
                            record.letter_annotations['cluster'][i] = last_cluster

                data[j] = (last_cluster, ini, end, size)
                last_cluster += 1
    return


def save_pymol_script(align: Bio.Align.MultipleSeqAlignment, pml_file: str, palette='Dark2', colors=7) -> None:
    """
         Cluster the alignments by structural similarity
         :param align: Protein alignment
         :type align: Bio.Align.MultipleSeqAlignment
         :param pml_file: Pymol script file name (pml)
         :type pml_file: str
         :param palette: colour palette
         :type palette: str
         :param colors: number of colours in the palette
         :type colors: int
    """
    data, idx = get_idx(align)

    tags = []
    pal = sns.color_palette(palette, colors).as_hex()
    with open(f'{pml_file}.pml', 'w') as f:
        f.write('# Script generated by Melodia\n\n')
        f.write('# load structures\n')
        for record in align:
            if 'structure' not in record.description:
                continue
            tags.append(record.id)
            f.write(f'load {record.id}.pdb\n')
        f.write('\n')

        f.write('# superimpose structures\n')
        for i in range(1, len(tags)):
            f.write(f'super {tags[i]}, {tags[0]}\n')
        f.write('\n')

        f.write('# non-conserved cluster color\n')
        f.write(f'color gray40\n\n')
        f.write('# cluster colors\n')
        for i in idx:
            positions = data[i]
            cluster, ini, end, size = positions
            if cluster >= 0:
                for record in align:
                    if 'structure' not in record.description:
                        continue
                    if record.letter_annotations['cluster'][ini] == cluster:
                        color = f'0x{pal[cluster % colors][1:]}'
                        f.write(f'color {color}, {record.id} and resi {ini + 1}-{end + 1}\n')
                f.write('\n')
        f.write('\ncenter\n')
    return


def get_idx(align: Bio.Align.MultipleSeqAlignment) -> Tuple[List[Tuple[Any, int, int, int]], List[SupportsIndex]]:
    """
    Internal function for clustering
    :param align: Protein alignment
    :return: Indexed clusters
    """
    clusters = {}
    for i, record in enumerate(align):
        if 'structure' not in record.description:
            continue
        for j, cluster in enumerate(record.letter_annotations['cluster']):
            if cluster not in clusters:
                clusters[cluster] = [j]
            else:
                clusters[cluster].append(j)
    data = []
    block_init = []
    for key, value in clusters.items():
        data.append((key, min(value), max(value), max(value) - min(value) + 1))
        block_init.append(min(value))
    idx = sorted(range(len(block_init)), key=block_init.__getitem__)
    return data, idx


def save_align_to_ps(align: Bio.Align.MultipleSeqAlignment, ps_file: str, palette='Dark2', colors=7) -> None:
    """
         Cluster the alignments by structural similarity
         :param align: Protein alignment
         :type align: Bio.Align.MultipleSeqAlignment
         :param ps_file: post-script file name
         :type ps_file: str
         :param palette: colour palette
         :type palette: str
         :param colors: number of colours in the palette
         :type colors: int
    """

    pal = sns.color_palette(palette, colors)
    rgb = [f'{c[0]:4.2f} {c[1]:4.2f} {c[2]:4.2f}' for c in pal]

    black = '0.00 0.00 0.00'
    grey = '0.50 0.50 0.50'

    out_file = f'{ps_file}.ps'
    with open(out_file, 'w') as ps:
        ps.write('%%!PS-Adobe-3.0\n')
        ps.write('%%%%Pages: 1\n')
        ps.write('%%%%Creator: Melodia 1.0\n')
        ps.write('%%%%CreationDate:\n')
        ps.write('%%%%EndComments\n')
        ps.write('%%%%Page: 1 1\n')
        ps.write('/Courier-Regular findfont  16.0 scalefont  setfont\n')
        ps.write('0.00 0.00 0.83 setrgbcolor\n')
        ps.write(f'72.0 735.0 moveto ({ps_file}) show\n')

        length = align.get_alignment_length()
        count = len(align)
        total = int(length / 50)
        block = count + 4
        blocks_per_page = int(76 / block) + 1

        page = 1
        line = 705.0
        blocks = 0
        position = 10

        for j in range(total + 1):
            ini = 0 + 50 * j
            end = 49 + 50 * j

            if end >= length:
                end = length - 1

            column = 203.0

            ps.write('/Courier-Regular findfont  8.0 scalefont  setfont\n')
            ps.write('0.00 0.00 0.83 setrgbcolor\n')

            for k in range(1, 6):
                if position < 100:
                    ps.write('%5.1f %5.1f moveto (%d) show\n' % (column, line, position))
                else:
                    ps.write('%5.1f %5.1f moveto (%d) show\n' % (column - 2.0, line, position))

                position += 10
                column += 80.0

                if position > length:
                    break

            i = 1
            for record in align:
                column = 52.0
                line -= 10.0
                ps.write('/Courier-Regular findfont  10.0 scalefont  setfont\n')
                ps.write('0.00 0.00 0.00 setrgbcolor\n')
                ps.write('%5.1f %5.1f moveto (%3d) show\n' % (column, line, i))

                column += 20.0

                ps.write('%5.1f %5.1f moveto (%s) show\n' % (column, line, record.id))

                column = 131.1
                bold = False
                last_color = black
                for cur_res in range(ini, end + 1):
                    if record.seq[cur_res] == '-':
                        if bold:
                            bold = False
                            ps.write('/Courier-Regular findfont  10.0 scalefont  setfont\n')
                        if last_color != black:
                            last_color = black
                            ps.write(f'{last_color} setrgbcolor\n')
                    else:
                        if 'structure' in record.description:
                            k = record.letter_annotations['cluster'][cur_res]
                            if k >= 0:
                                c = k % colors
                                if not bold:
                                    bold = True
                                    ps.write('/Courier-Bold findfont  10.0 scalefont  setfont\n')
                                if last_color != rgb[c]:
                                    last_color = rgb[c]
                                    ps.write(f'{last_color} setrgbcolor\n')
                            else:
                                if bold:
                                    bold = False
                                    ps.write('/Courier-Regular findfont  10.0 scalefont  setfont\n')
                                if last_color != grey:
                                    last_color = grey
                                    ps.write(f'{last_color} setrgbcolor\n')
                        else:
                            if bold:
                                bold = False
                                ps.write('/Courier-Regular findfont  10.0 scalefont  setfont\n')
                            if last_color != black:
                                last_color = black
                                ps.write(f'{last_color} setrgbcolor\n')
                    if 'structure' in record.description:
                        if record.letter_annotations['cluster'][cur_res] < 0:
                            ps.write(f'%5.1f %5.1f moveto (%c) show\n' % (column, line, record.seq[cur_res].lower()))
                        else:
                            ps.write(f'%5.1f %5.1f moveto (%c) show\n' % (column, line, record.seq[cur_res]))
                    else:
                        ps.write(f'%5.1f %5.1f moveto (%c) show\n' % (column, line, record.seq[cur_res].lower()))
                    column += 8.0
                i += 1

            line -= 20.0
            blocks += 1

            if blocks == blocks_per_page and j != total:
                blocks = 0
                line = 705.0
                page += 1

                ps.write('showpage\n')
                ps.write(f'%%%%Page: {page} {page}\n')
                ps.write('/Courier-Regular findfont  16.0 scalefont  setfont\n')
                ps.write('0.00 0.00 0.83 setrgbcolor\n')
                ps.write(f'72.0 735.0 moveto (out_file) show\n')

        ps.write('showpage\n')

    return



# Copyright 2021-2024 KU Leuven.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Rinaldo Wander Montalvão, PhD
#
import math
import numpy as np

from Bio.PDB import Chain
from typing import Dict, List, Tuple
from collections import defaultdict
from dataclasses import dataclass, field
from numpy.polynomial import chebyshev
from scipy.interpolate import CubicSpline


@dataclass
class ResidueGeometry:
    # Residue information
    name: str = ""
    chain: str = ""
    res_num: int = 0
    res_order: int = 0

    # Differential geometry
    curvature: float = 0.0
    torsion: float = 0.0
    arc_len: float = 0.0

    # Knot theory invariant
    writhing: float = 0.0

    # Dihedral angles
    phi: float = 0.0
    psi: float = 0.0

    # Residue annotation
    res_ann: Dict[str, str] = field(default_factory=lambda: defaultdict(dict))

    # Custom float value
    custom: float = 0.0


class GeometryParser:
    """
    Class for parsing the geometrical properties of a protein chain
    """
    __slots__ = ('__residues',
                 '__residues_map',
                 '__degrees',
                 '__gap_list',
                 '__anomaly_list')

    def __init__(self, chain: Chain, deg: bool = True) -> None:
        """
        :param chain: Protein chain
        :type chain: Chain
        :param deg: Degree?
        :type deg: bool
        """
        residues, residues_map = GeometryParser.calc_geometry(chain=chain, deg=deg)
        self.__residues = residues
        self.__residues_map = residues_map
        self.__degrees = deg
        self.__gap_list = GeometryParser.find_gaps(chain=chain)
        self.__anomaly_list = GeometryParser.find_anomalies(chain=chain)

    @property
    def residues(self) -> Dict[str, ResidueGeometry]:
        """
        Access to residue geometry
        :return: dataclass for residue geometry
        :rtype: dict[int, ResidueGeometry]
        """
        return self.__residues

    @property
    def residues_map(self) -> Dict[str, ResidueGeometry]:
        """
        Maps residues number
        :return: Dictionary for the residue map
        :rtype: dict[int, ResidueGeometry]
        """
        return self.__residues_map

    @property
    def deg(self) -> bool:
        """
        Are phi and psi in degrees?
        :return: True if phi and psi are in degrees, False for radians
        :rtype: bool
        """
        return self.__degrees

    @staticmethod
    def find_gaps(chain: Chain) -> List[Tuple[int, int]]:
        """
        Find gaps in the protein's chain
        :param chain: Protein chain
        :type chain: Chain
        :return: List of gaps
        :rtype: list[tuple[int, int]]
        """
        gaps = []

        i = 0
        residue = list(chain.get_residues())[i]
        het_flag, prev, insertion_code = residue.id

        for i in range(len(chain)):
            # get chain and pos
            residue = list(chain.get_residues())[i]
            het_flag, pos, insertion_code = residue.id

            if pos - prev > 1:
                gaps.append((prev, pos))

            prev = pos

        return gaps

    @property
    def gaps(self) -> List[Tuple[int, int]]:
        """
        Access to chain's gaps
        :return:  Chain's gaps
        :rtype: list[tuple[int, int]]
        """
        return self.__gap_list

    @property
    def gap(self) -> bool:
        """
        Are there  gaps in the chain?
        :return: True if gaps were found
        :rtype: bool
        """
        return len(self.__gap_list) > 0

    @staticmethod
    def find_anomalies(chain: Chain) -> List[str]:
        """
        Find anomalies in the protein's chain
        :param chain: Protein chain
        :type chain: Chain
        :return: List of anomalies
        :rtype: list[str]
        """
        # TODO: insert anomalies
        anomalies = []

        return anomalies

    @property
    def anomalies(self) -> List[str]:
        """
        Access to chain's anomalies
        :return:  Chain's anomalies
        :rtype: list[str]
        """
        return self.__anomaly_list

    @property
    def anomaly(self) -> bool:
        """
        Are there anomalies in the chain?
        :return: True if anomalies were found
        :rtype: bool
        """
        return len(self.__anomaly_list) > 0

    @staticmethod
    def calc_curvature_torsion(p: float,
                               t: List[float],
                               xt: CubicSpline,
                               yt: CubicSpline,
                               zt: CubicSpline) -> (float, float):
        """
        Function to compute Curvature and Torsion
        :param p: point of calculation
        :type p: float
        :param t: list os parameters
        :type t: list[float]
        :param xt: x(t)
        :type xt: CubicSpline
        :param yt: y(t)
        :type yt: CubicSpline
        :param zt: z(t)
        :type zt: CubicSpline
        :return: curvature, torsion
        :rtype: (float, float)
        """
        number_points = 51

        # TODO: Improve balance method
        mn = np.min(t)
        mx = np.max(t)
        cxt, result_xt = 0.0, 0.0
        cyt, result_yt = 0.0, 0.0
        czt, result_zt = 0.0, 0.0
        for dt in range(1, 4):
            delta = float(dt)

            ini = p - delta
            end = p + delta

            if ini < mn:
                offset = mn - ini
            elif end > mx:
                offset = mx - end
            else:
                offset = 0.0

            ini += offset
            end += offset

            tp = np.linspace(ini, end, number_points)

            cxt, result_xt = chebyshev.chebfit(tp, xt(tp), deg=10, full=True)
            cyt, result_yt = chebyshev.chebfit(tp, yt(tp), deg=10, full=True)
            czt, result_zt = chebyshev.chebfit(tp, zt(tp), deg=10, full=True)

            if result_xt[0].size != 0 and result_yt[0].size != 0 and result_zt[0].size != 0:
                break

        cxt_d1 = chebyshev.chebder(cxt, m=1)
        cyt_d1 = chebyshev.chebder(cyt, m=1)
        czt_d1 = chebyshev.chebder(czt, m=1)

        cxt_d2 = chebyshev.chebder(cxt, m=2)
        cyt_d2 = chebyshev.chebder(cyt, m=2)
        czt_d2 = chebyshev.chebder(czt, m=2)

        cxt_d3 = chebyshev.chebder(cxt, m=3)
        cyt_d3 = chebyshev.chebder(cyt, m=3)
        czt_d3 = chebyshev.chebder(czt, m=3)

        xt_d1 = chebyshev.chebval(p, cxt_d1)
        yt_d1 = chebyshev.chebval(p, cyt_d1)
        zt_d1 = chebyshev.chebval(p, czt_d1)

        xt_d2 = chebyshev.chebval(p, cxt_d2)
        yt_d2 = chebyshev.chebval(p, cyt_d2)
        zt_d2 = chebyshev.chebval(p, czt_d2)

        xt_d3 = chebyshev.chebval(p, cxt_d3)
        yt_d3 = chebyshev.chebval(p, cyt_d3)
        zt_d3 = chebyshev.chebval(p, czt_d3)

        # Compute curvature
        v1 = np.array([xt_d1, yt_d1, zt_d1])
        v2 = np.array([xt_d2, yt_d2, zt_d2])

        rs = np.cross(v1, v2)
        r1 = np.dot(rs, rs)
        r2 = np.dot(v1, v1)

        curvature = math.sqrt(r1) / math.sqrt(r2) ** 3

        # Compute torsion
        det = -xt_d3 * yt_d2 * zt_d1
        det += xt_d2 * yt_d3 * zt_d1
        det += xt_d3 * yt_d1 * zt_d2
        det -= xt_d1 * yt_d3 * zt_d2
        det -= xt_d2 * yt_d1 * zt_d3
        det += xt_d1 * yt_d2 * zt_d3

        torsion = det / r1

        return curvature, torsion

    @staticmethod
    def calc_arc_length(p: float, xt: CubicSpline, yt: CubicSpline, zt: CubicSpline) -> float:
        """
        Compute the arc length of a 3-residues long curve
        :param p: point around the curve is calculated
        :type p: float
        :param xt: x(t)
        :type xt: CubicSpline
        :param yt: y(t)
        :type yt: CubicSpline
        :param zt: z(t)
        :type zt: CubicSpline
        :return: arc length
        :rtype: float
        """
        arc_len = 0.0

        i = p - 1.0
        while i < (p + 1.0):
            dx = xt(i + 0.1) - xt(i)
            dy = yt(i + 0.1) - yt(i)
            dz = zt(i + 0.1) - zt(i)

            dist = np.array([dx, dy, dz])

            arc_len += math.sqrt(np.dot(dist, dist))

            i += 0.1

        return arc_len

    @staticmethod
    def calc_writhing(i: int, t: List[float], x: List[float], y: List[float], z: List[float]) -> float:
        """
        Compute the writhing number in a 5-residue long window
        :param i: residue postion
        :type i: int
        :param t: curve's parameters
        :type t: list[float]
        :param x: x(t)
        :type x: list[float]
        :param y: y(t}
        :type y: list[float]
        :param z: z(t)
        :type z: list[float]
        :return: writhing number
        :rtype: float
        """
        start = i - 2
        stop = i + 2

        ini = 0
        end = len(t) - 1
        if start < ini:
            offset = ini - start
        elif stop > end:
            offset = end - stop
        else:
            offset = 0

        start += offset
        stop += offset

        rij = np.zeros(3)
        ri1j = np.zeros(3)
        rij1 = np.zeros(3)
        rjj1 = np.zeros(3)
        rii1 = np.zeros(3)
        ri1j1 = np.zeros(3)

        # Return the number's sign
        def sgn(v: np.ndarray) -> float:
            return v and (1.0, -1.0)[v < 0.0]

        total = 0.0
        for ii in range(start, stop - 2):
            for jj in range(ii + 2, stop):
                rij[0] = x[jj] - x[ii]
                rij[1] = y[jj] - y[ii]
                rij[2] = z[jj] - z[ii]

                ri1j[0] = x[jj] - x[ii + 1]
                ri1j[1] = y[jj] - y[ii + 1]
                ri1j[2] = z[jj] - z[ii + 1]

                rij1[0] = x[jj + 1] - x[ii]
                rij1[1] = y[jj + 1] - y[ii]
                rij1[2] = z[jj + 1] - z[ii]

                ri1j1[0] = x[jj + 1] - x[ii + 1]
                ri1j1[1] = y[jj + 1] - y[ii + 1]
                ri1j1[2] = z[jj + 1] - z[ii + 1]

                rjj1[0] = x[jj + 1] - x[jj]
                rjj1[1] = y[jj + 1] - y[jj]
                rjj1[2] = z[jj + 1] - z[jj]

                rii1[0] = x[ii + 1] - x[ii]
                rii1[1] = y[ii + 1] - y[ii]
                rii1[2] = z[ii + 1] - z[ii]

                aij = (np.cross(rij, rij1) / np.linalg.norm(np.cross(rij, rij1)))
                bij = (np.cross(rij1, ri1j1) / np.linalg.norm(np.cross(rij1, ri1j1)))
                cij = (np.cross(ri1j1, ri1j) / np.linalg.norm(np.cross(ri1j1, ri1j)))
                dij = (np.cross(ri1j, rij) / np.linalg.norm(np.cross(ri1j, rij)))

                omegaij = (math.asin(np.dot(aij, bij)) +
                           math.asin(np.dot(bij, cij)) +
                           math.asin(np.dot(cij, dij)) +
                           math.asin(np.dot(dij, aij))) * sgn(np.dot(np.cross(rjj1, rii1), rij1))

                total += omegaij / (4.0 * math.pi)
        writhing = 2.0 * total
        return writhing

    @staticmethod
    def calc_geometry(chain: Chain, deg: bool) -> Dict[int, ResidueGeometry]:
        """
        Function used to compute the geometric properties around residues.
        It computes curvature, torsion, arc-length and writhing number
        :param chain: Protein main-chain
        :type chain: Chain
        :param deg: angle in degrees?
        :type deg: bool
        :return:  Residue dictionary
        :rtype: Dict[int, ResidueGeometry]
        """
        t = []
        x = []
        y = []
        z = []

        residues_map = {}

        residues = {}
        num = 0
        for residue in chain:
            # Skip invalid residues
            res_type, model, chain_id, res_id = residue.get_full_id()
            het_flag, pos, insertion_code = res_id
            if het_flag[0] != " ":
                continue

            try:
                coord = residue["CA"].get_coord()
            except KeyError:
                print(f'ERROR: missing CA atom at {residue.get_resname()} - {residue.get_full_id()}!')
                raise

            t.append(float(num))
            x.append(coord[0])
            y.append(coord[1])
            z.append(coord[2])

            residues[num] = ResidueGeometry(name=residue.get_resname(),
                                                 chain=chain_id,
                                                 res_num=num,
                                                 res_order=pos,
                                                 curvature=0.0,
                                                 torsion=0.0,
                                                 arc_len=0.0,
                                                 writhing=0.0,
                                                 res_ann={},
                                                 custom=0.0)
            residues_map[pos] = num

            num += 1

        # Fit the alpha-carbons with a cubic-spline
        xt = CubicSpline(t, x, bc_type='natural')
        yt = CubicSpline(t, y, bc_type='natural')
        zt = CubicSpline(t, z, bc_type='natural')

        ini = 0
        end = len(t) - 1
        for i, tp in enumerate(t):
            # Compute curvature and torsion
            if ini < i < end:
                curvature, torsion = GeometryParser.calc_curvature_torsion(p=tp, t=t, xt=xt, yt=yt, zt=zt)
            elif i == ini:
                curvature, torsion = GeometryParser.calc_curvature_torsion(p=t[+1], t=t, xt=xt, yt=yt, zt=zt)
            else:
                curvature, torsion = GeometryParser.calc_curvature_torsion(p=t[-2], t=t, xt=xt, yt=yt, zt=zt)

            # Compute the arc length
            arc_len = GeometryParser.calc_arc_length(p=tp, xt=xt, yt=yt, zt=zt)

            # Compute the writhing number
            writhing = GeometryParser.calc_writhing(i=i, t=t, x=x, y=y, z=z)

            residues[int(tp)].curvature = curvature
            residues[int(tp)].torsion = torsion
            residues[int(tp)].arc_len = arc_len
            residues[int(tp)].writhing = writhing

        GeometryParser.calc_dihedral_angles(chain=chain, residues=residues, deg=deg)

        return residues, residues_map

    @staticmethod
    def calc_dihedral_torsion(p1: np.ndarray, p2: np.ndarray, p3: np.ndarray, p4: np.ndarray, deg: bool) -> float:
        """
        Compute the dihedral angles between four vectors
        :param p1: vector 1
        :type p1: np.ndarray
        :param p2: vector 2
        :type p2:  np.ndarray
        :param p3: vector 3
        :type p3: np.ndarray
        :param p4: vector 4
        :type p4: np.ndarray
        :param deg: dihedral angle
        :type deg: bool
        :return: Dihedral angle (in degrees if deg=True, radians otherwise)
        :rtype: float
        """
        b1 = p2 - p1
        b2 = p2 - p3
        b3 = p4 - p3

        # Normalize a vector
        def norm_vec(v: np.ndarray) -> np.ndarray:
            return v/np.linalg.norm(v)

        n1 = norm_vec(np.cross(b1, b2))
        n2 = norm_vec(np.cross(b2, b3))

        m1 = np.cross(n1, norm_vec(b2))

        x = np.dot(n1, n2)
        y = np.dot(m1, n2)

        if deg:
            theta = math.degrees(math.atan2(y, x))
        else:
            theta = math.atan2(y, x)
        return theta

    @staticmethod
    def calc_dihedral_angles(chain: Chain, residues: Dict[int, ResidueGeometry], deg: bool) -> None:
        """
        Compute the dihedral angles phi and psi.
        :param chain: Protein chain
        :type chain: Chain
        :param residues:  Residue dictionary
        :type residues: Dict[int, ResidueGeometry]
        :param deg: angle in degrees?
        :type deg: bool
        """
        idx = [res.id[1] for res in list(chain.get_residues()) if res.id[0] == ' ']

        ini = idx[+0]
        end = idx[-2]

        num = 0
        for i, residue in enumerate(chain):
            # Skip invalid residues
            het_flag, pos, insertion_code = residue.id
            if het_flag[0] != ' ':
                continue

            # core atoms
            try:
                atom_n = residue['N'].get_coord()
                atom_ca = residue['CA'].get_coord()
                atom_c = residue['C'].get_coord()
            except KeyError:
                pdb, model, chain_id = chain.full_id
                print(f'Error: Missing N, CA or C atom at [{pos}] {pdb} - {model} - {chain_id}')
                raise

            # phi
            if pos > ini:
                het_flag, prev_res, insertion_code = chain[idx[i - 1]].id

                if abs(prev_res - pos) <= 1:
                    try:
                        p1 = chain[idx[i - 1]]['C'].get_coord()
                    except KeyError:
                        pdb, model, chain_id = chain.full_id
                        print(f'Error: Missing C atom at [{idx[i - 1]}] {pdb} - {model} - {chain_id}')
                        raise
                    p2 = atom_n
                    p3 = atom_ca
                    p4 = atom_c
                    phi = GeometryParser.calc_dihedral_torsion(p1=p1, p2=p2, p3=p3, p4=p4, deg=deg)
                else:
                    phi = 0.0
            else:
                phi = 0.0

            # psi
            if pos < end:
                try:
                    het_flag, next_res, insertion_code = chain[idx[i + 1]].id
                except (IndexError, KeyError):
                    pdb, model, chain_id = chain.full_id
                    print(f'Error: after residue [{idx[i]}] {pdb} - {model} - {chain_id}')
                    raise

                if abs(next_res - pos) <= 1:
                    p1 = atom_n
                    p2 = atom_ca
                    p3 = atom_c
                    try:
                        p4 = chain[idx[i + 1]]['N'].get_coord()
                    except KeyError:
                        pdb, model, chain_id = chain.full_id
                        print(f'Error: Missing N atom at [{idx[i + 1]}] {pdb} - {model} - {chain_id}')
                        raise
                    psi = GeometryParser.calc_dihedral_torsion(p1=p1, p2=p2, p3=p3, p4=p4, deg=deg)
                else:
                    psi = 0.0
            else:
                psi = 0.0

            residues[num].phi = phi
            residues[num].psi = psi

            num += 1



from .melodia_py import geometry_from_structure_file, geometry_dict_from_structure, dataframe_from_alignment
from .melodia_py import geometry_from_structure, bfactor_from_geo, view_putty, view_cartoon, view_tube
from .melodia_py import PropensityTable, parser_pir_file, cluster_alignment, save_align_to_ps, save_pymol_script


__version__ = '0.1.0'
__author__ = 'Rinaldo Wander Montalvão, PhD'



import setuptools

setuptools.setup(
    name="melodia_py",
    version="0.1.4",
    url="https://github.com/rwmontalvao/Melodia_py",
    author="Rinaldo Wander Montalvão, PhD",
    author_email="rwmontalvao@gmail.com",
    description="Differential Geometry of Proteins Backbones",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    install_requires=['biopython>=1.83',
                      'pandas>=2.2.2',
                      'typing-extensions>=4.12',
                      'seaborn>=0.13.2',
                      'jupyterlab>=4.2.2',
                      'tqdm>=4.66.4',
                      'dill>=0.3.8',
                      'nglview>=3.1.2',
                      'pyarrow>=16.1.0',
                      'scikit-learn>=1.5.0',
                      'sty>=1.0.0',
                      'more-itertools>=10.3.0'],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.12',
    ],
    include_package_data=True,
)
